/* [ NETSCRIPT: lightweight tcp socket scripting -- version 1.5.7 ] **********
 * (portable/multi-platform) lightweight tcp socket scripting.  intended for *
 * (non-)experienced persons to be able to use to automate situations, built *
 * on a word-to-word ruleset response system.  includes wildcard support,    *
 * character replacement, random replacement, argument inclusion, server     *
 * timeout, initial send, display altering, multiple character dump formats, *
 * telnet protocol support, logging, program to socket dumping, executable   *
 * ruleset support and reverse binding among other things.                   *
 *                                                                           *
 * AUTHOR:                                                                   *
 *  v9/vade79@realhalo.org, realhalo.org. (* version)                        *
 *                                                                           *
 * COMPILE:                                                                  *
 *  system type(generic):                                                    *
 *   # cc netscript.c -o netscript                                           *
 *                                                                           *
 *  system type(Linux):                                                      *
 *   # cc netscript.c -o netscript -DLINUX                                   *
 *                                                                           *
 *  system type(BSD):                                                        *
 *   # cc netscript.c -o netscript -DBSD                                     *
 *                                                                           *
 *  system type(IRIX):                                                       *
 *   # cc netscript.c -o netscript -DIRIX                                    *
 *                                                                           *
 *  system type(SunOS/Solaris):                                              *
 *   # cc netscript.c -o netscript -lnsl -lsocket -DSUNOS                    *
 *                                                                           *
 * BASIC TUTORIAL:                                                           *
 *  the minimal command line requires -r(or -s), -h(or -b) and -p, like so:  *
 *                                                                           *
 *   # ./netscript -r rulesetfile -h host.com -p port                        *
 *                                                                           *
 *  or, if you have it properly set up(like examples that come with the      *
 *  package) you can run it like so:                                         *
 *                                                                           *
 *   # chmod +x ./rulesetfile;./rulesetfile                                  *
 *                                                                           *
 *  the ruleset file(-r argument) should contain a list similar to this:     *
 * ------------------------------------------------------------------------- *
 *  001 Started remote daemon.                                               *
 *  USER test                                                                *
 *  002 Enter password.                                                      *
 *  PASS test                                                                *
 * ------------------------------------------------------------------------- *
 *  this ruleset file would send "USER test" upon receiving the data "001    *
 *  Started remote daemon." and "PASS test" upon receiving the data "002     *
 *  Enter password."                                                         *
 *                                                                           *
 * EXTENDED TUTORIAL:                                                        *
 *  using variables and wildcard matches are also used, taken from the basic *
 *  tutorial ruleset file with some modifications:                           *
 * ------------------------------------------------------------------------- *
 *  001 $*                                                                   *
 *  USER test                                                                *
 *  002 $*                                                                   *
 *  PASS test                                                                *
 * ------------------------------------------------------------------------- *
 *  this would send the same data as in the basic tutorial.  here is a list  *
 *  of the wild card match variables, for use only with the output of the    *
 *  ruleset(use the -d option of netscript for more specifics):              *
 *                                                                           *
 *   $*     = if words before match until the point this variable is used.   *
 *   $?     = if anything fills a single word.                               *
 *   $ALPHA = if the word is an alphabetical word.                           *
 *   $DIGIT = if the word is a numerical word.                               *
 *   $ALNUM = if the word is an alphabetical+numerical word.                 *
 *   $UPPER = if the word is an upper case word.                             *
 *   $LOWER = if the word is a lower case word.                              *
 *   $PUNCT = if the word is a punctuated word.                              *
 *   $CNTRL = if the word is a control character word.                       *
 *   $PRINT = if the word is a printable word.                               *
 *   $####  = must be 4 digits exactly(# = numeric).  if the word is equal   *
 *            in length to the value supplied. (for example: checking for a  *
 *            length equal to 12 would be $0012)                             *
 *   $@     = not truly a wildcard, must be by itself.  stops using rules.   *
 *   $!     = not truly a wildcard, must be by itself.  closes the socket.   *
 *   $~     = not truly a wildcard, must be by itself.  asks data to send.   *
 *   $%     = not truly a wildcard, must be by itself.  displays text.       *
 *   $_     = not truly a wildcard, must be by itself.  sends without CR.    *
 *   $-     = not truly a wildcard, must be by itself.  uses rule only once. *
 *   $<     = dumps the supplied file after the variable to the socket.      *
 *   $>     = writes the info that matched to the supplied file after the    *
 *            variable.                                                      *
 *   $+     = appends the info that matched to the supplied file after the   *
 *            variable.                                                      *
 *   $=     = dumps the execution data of the supplied file after the        *
 *            variable to the socket.                                        *
 *   $##### = must be 5 digits exactly(# = numeric).  this variable must be  *
 *            tagged on the end of an output line(the last data on the line) *
 *            to work, it sleeps the time of the value supplied. (for        *
 *            example: $00012 would sleep 12 seconds)                        *
 *                                                                           *
 *  note: anything attached after for $@, $!, $~, $%, $_, $<, $>, $+, $= and *
 *  $- will be used for either text/display, information or a filename.      *
 *  dependant on the type of variable being used.  (in $-'s case, the rule   *
 *  to be used one time.  $- requires data after it to be taken as a         *
 *  variable)                                                                *
 *                                                                           *
 *  now, here is an example of a numerical variable in a new ruleset:        *
 * ------------------------------------------------------------------------- *
 *  001 Daemon ready.                                                        *
 *  USER test                                                                *
 *  002 Sorry, user $? needs a password.                                     *
 *  USER $3 mypasswd                                                         *
 * ------------------------------------------------------------------------- *
 *  as you can see, once the ruleset sees "001 Daemon ready." it sends "USER *
 *  test", but the format is USER <user> <pass> on this daemon.  so, upon    *
 *  receiving the error it sends the 3rd argument from the error message as  *
 *  the <user> argument, which should be the same username.  this is not     *
 *  practical, but an example.  here are the numerical/random variables, for *
 *  use with all aspects of the ruleset(use the -d option of netscript for   *
 *  more specifics):                                                         *
 *                                                                           *
 *   $1-9     = given argument responses, from the server.                   *
 *   $01-FF   = hex->character replacement.                                  *
 *   $001-255 = dec->character replacement.                                  *
 *   $&       = random alphabetical character replacement.                   *
 *   $#       = random numerical character replacement.                      *
 *                                                                           *
 * ENVIRONMENT:                                                              *
 *  $NS_CMDLINE                                                              *
 *   this environmental variable will take the data in the variable as a     *
 *   command line, it has priority over typical command line usage.  but,    *
 *   will not override +x files.                                             *
 *                                                                           *
 *  $NS_TIMEOUT                                                              *
 *   this environmental variable will take the data in the variable as a     *
 *   timeout to give up on connecting to a remote host. (25 seconds is the   *
 *   default if no environmental variable is used)                           *
 *                                                                           *
 *  $COLUMNS                                                                 *
 *   this environmental variable will take the data in the variable as a     *
 *   maximum number of characters to print per line for input/output         *
 *   displaying.                                                             *
 *                                                                           *
 * COMMENTS:                                                                 * 
 *  like most things, i perfer things to be portable.  so i made this around *
 *  a single file.  other files are not needed for netscript to run          *
 *  properly.  i tried to make netscript as basic as possible, with minimal  *
 *  usage of uncommon functions, so it can be easily portable.  platform to  *
 *  platform.  if you do not understand the proper way to make a ruleset,    *
 *  you may want to consult the packaged netscript(if this is not already    *
 *  from the packaged netscript) for examples of usage.                      *
 *                                                                           *
 * netscript.c: program source code for netscript.      (1666l!7120w!60852b) *
 *****************************************************************************/
#include <netdb.h> /* socket related. */
#include <netinet/in.h> /* socket related. */
#include <signal.h> /* to handle signals.  ctrl-c, memory errors, etc. */
#include <stdio.h> /* multiple uses. */
#include <stdlib.h> /* multiple uses. */
#include <string.h> /* multiple uses. */
#include <sys/time.h> /* log times. */
#include <sys/types.h> /* socket related. */
#include <sys/socket.h> /* socket related. */
#include <unistd.h> /* multiple uses. */
/* definitions, most can be changed.  but, not particularly recommended. */
#define VERSION "1.5.7" /* version information. */
#define MAX_ARGS 1024 /* total input/output slots allowed, combined. */
#define BASE_BUFFER 1024 /* generic buffer size, base for all increments. */
#define DFL_COLUMNS 80 /* default characters per line with in/out data. */
#define PARAMETER_VAR_CHAR 0x24 /* identifier for args/convs/len checks, $. */
#define ENV_CMDLINE "NS_CMDLINE" /* env var treated as a command line. */
#define ENV_TIMEOUT "NS_TIMEOUT" /* env var for connection (to) timeout. */
#define ENV_COLUMNS "COLUMNS" /* env var for in/out data line size. */
#define NS_ANY "$?" /* wildcard for any word. */
#define NS_ALL "$*" /* wildcard for any words after that point. */
#define NS_ALPHA "$ALPHA" /* wildcard for alphabet characters. */
#define NS_DIGIT "$DIGIT" /* wildcard for numeric characters. */
#define NS_ALNUM "$ALNUM" /* wildcard for alphabet and numeric characters. */
#define NS_LOWER "$LOWER" /* wildcard for lower case words. */
#define NS_UPPER "$UPPER" /* wildcard for upper case words. */
#define NS_PUNCT "$PUNCT" /* wildcard for punctuated words. */
#define NS_CNTRL "$CNTRL" /* wildcard for control character words. */
#define NS_PRINT "$PRINT" /* wildcard for printable words. */
#define NS_STOP "$@" /* pre-match, disregards ruleset after this. */
#define NS_QUIT "$!" /* pre-match, exits netscript if exact. */
#define NS_ASK "$~" /* pre-match, asks data to reply with. */
#define NS_ECHO "$%" /* pre-match, displays data after variable. */
#define NS_RAW "$_" /* pre-match, writes to the socket without \n. */
#define NS_DUMP "$<" /* pre-match, dumps the following file to the socket. */
#define NS_WRITE "$>" /* pre-match, writes the match line to a file. */
#define NS_APPEND "$+" /* pre-match, appends the match line to a file. */
#define NS_EXEC "$=" /* pre-match, locally executes a file. */
#define NS_ONCE "$-" /* pre-match, uses rule one time per connection. */
#define INPUTPATH "/dev/tty" /* path to read for standard input. */
/* checks for defines -- if never get defined, puts generic. */
/* generated to linux. */
#ifdef LINUX
#define COMPTYPE "lin-gen"
#endif
/* generated to bsd. */
#ifdef BSD
#undef COMPTYPE
#define COMPTYPE "bsd-gen"
#endif
/* generated to sunos. */
#ifdef SUNOS
#undef COMPTYPE
#define COMPTYPE "sns-gen"
#endif
/* generated to IRIX. */
#ifdef IRIX
#undef COMPTYPE
#define COMPTYPE "irx-gen"
#endif
/* no generation. */
#ifndef COMPTYPE
#define COMPTYPE "generic"
#endif
/* these should not be changed at all, space savers.  more so, made to save */ 
/* me typing time.  i am lazy.  if anyone plans on reading this code other */
/* than myself, i am sorry.  but, filesize and my laziness comes first. */
#define A alarm
#define AT atoi
#define B bzero
#define BI bind
#define BR break
#define C char
#define CL close
#define CO connect
#define CS case
#define E else
#define EC extern char
#define EI else if
#define F for
#define FC fclose
#define FG fgetc
#define FL FILE
#define FO fopen
#define FP fprintf
#define FR free
#define FRO freopen
#define FS fgets
#define I int
#define IF if /* so i do not mix ucase and lcase, obviously not for space. */
#define M malloc
#define P printf
#define PC pclose
#define R return
#define RD read
#define S struct
#define SA sockaddr
#define SAN sockaddr_in
#define SC strcmp
#define SCA strcat
#define SCC strcasecmp
#define SCP strcpy
#define SD shutdown
#define SDU strdup
#define SG signal
#define SH short
#define SI signed
#define SK socket
#define SL strlen
#define SLP sleep
#define SNC strncmp
#define SNCC strncasecmp
#define SO sizeof
#define SP sprintf
#define ST static
#define SW switch
#define T time_t
#define U unsigned
#define V void
#define W while
#define WR write
/* global stored information, used throughout.  also, id info. */
/* these are all filled with 0 or 1. (to that extent) */
U SH rs_static[((MAX_ARGS/2)+2)];
U SH rs_dynamic[((MAX_ARGS/2)+2)];
U SH bindmode=0;
U SH displaytype=0;
U SH initsend=0;
U SH inputrules=0;
U SH log=0;
U SH nofrom=0;
U SH noshowa=0;
U SH noshowc=0;
U SH noshowp=0;
U SH nosrecv=0;
U SH nossend=0;
U SH notnetopt=0;
U SH nowrap=0;
U SH printonly=0;
U SH runcmd=0;
U SH runexit=0;
U SH stoprules=0;
U SH tnet=0;
U SH truetnet=0;
/* these could end up being larger sets. */
U I rs_delay[((MAX_ARGS/2)+2)];
U I alrm=0;
U I columns=0;
U I tot_i=0;
U I tot_o=0;
U I xalrm=0;
I sock=0;
I ssock=0;
ST C id[]="$Id: netscript.c,v "VERSION" 2001/04/23 02:38:57 EST vade79 Exp $";
C *input[((MAX_ARGS/2)+2)];
C *output[((MAX_ARGS/2)+2)];
C *execfile;
C *iswrite;
C *logfile;
C *progname;
C *toenv;
C *parm;
C *parsedline;
C *swrite;
EC *optarg;
/* functions, mainly put this as an index.  since i am not returning */
/* anything except integers, i noticed that it takes less space to just */
/* return short true/false responses and write to global arrays if needed, */
/* especially for for routines used multiple times in different situations. */
/* (prototypes) */
V sighandler(I); /* all-round signal handler. */
V parseoutput(C *,C *); /* replace input arguments into output resp. */
V parsecharvars(C *); /* replace hex, dec and random variables. */
V makelists(C *); /* make in/out lists, from the ruleset. */
V pe(C *,U SH); /* prints data (and/or) exits. */
V pd(U C *,U SH,U I); /* prints in/out data, formats. */
V logdata(C *,U SH); /* adds data to the log file. */
V dumpexec(C *,I); /* execution of a program that dumps to the socket. */
V sockconnect(C *,U I); /* all-round socket handler. */
V showinfo(U SH); /* show different types of version information. */
V displayv(V); /* displays the variables for matches and replacements. */
V usage(U SH); /* displays the program usage. */
V nsexit(SH); /* optional exit routine to be placed. */
U SH wordmatch(C *,C *); /* wildcard match handler. */
U SH parameter(C *,I,U I); /* gets a parameter from the array. */
U SH wischeck(C *,U I); /* checks for word format of its counterpart. */
U SH getexecargs(C *); /* gets arguments from +x files. */
/* this handles the signals netscript receives. */
V sighandler(I sig){
 /* clean up before exiting completely. */
 IF(sock){
  SD(sock,2);
  sock=CL(sock);
 }
 IF(ssock){
  SD(ssock,2);
  ssock=CL(ssock);
 }
 SW(sig){
  CS SIGINT:
   P("(killed -- user aborted)\n");
   BR;
  CS SIGSEGV:
   pe("netscript failed in memory management, please report this event and"
   " how it occured",1);
   BR;
  CS SIGALRM:
   P("(killed -- alarm)\n");
   BR;
  default:
   P("(killed -- undefined signal: %d)\n",sig);
   BR;
 }
 nsexit(1);
}
/* this is the function that fills in the $0-9 replacement of output. */
/* the function parsecharvars() is used for both input and output of the */
/* ruleset file.  making these two functions somewhat differ, so they each */
/* got their own function. */
V parseoutput(C *input,C *output){
 U I i=0;
 C *tmpio;
 /* impossible to be larger than both combined. */
 IF(!(tmpio=(C *)M(SL(input)+SL(output)+1)))
  pe("parseoutput(): allocation of memory error",1);
 B(tmpio,(SL(input)+SL(output)+1));
 W(!parameter(output,i++,0x20)){
  IF(SL(parm)==2&&parm[0]==PARAMETER_VAR_CHAR&&isdigit(parm[1]))
   parameter(input,(parm[1]-0x30),0x20);
  IF(i!=1)
   SCA(tmpio," ");
  SCA(tmpio,parm);
 }
 FR(swrite);
 IF(!(swrite=(C *)SDU(tmpio)))
  pe("parseoutput(): duplication of memory error",1);
 FR(tmpio);
 R;
}
/* this function is meant to replace $001-255 and $01-FF to character */
/* format for the ruleset, also for replacement of the $& and $# random */
/* variables. (not server related) */
V parsecharvars(C *line){
 U I i=0;
 U I j=0;
 U I tmp=0;
 S timeval tv;
 FR(parsedline);
 IF(!(parsedline=(C *)M(SL(line)+1)))
  pe("parsecharvars(): allocation of memory error",1);
 F(i=0;i<SL(line);i++){
  IF(PARAMETER_VAR_CHAR==line[i]){
   /* handle decimal truncation. */
   IF(line[i+3]&&isdigit(line[i+1])&&isdigit(line[i+2])&&isdigit(line[i+3])){
   /* to not get confused with the length check wildcard. */
    IF(!(line[i+4]&&isdigit(line[i+4]))){
     tmp=(((line[i+1]-0x30)*100)+((line[i+2]-0x30)*10)+(line[i+3]-0x30));
     IF(tmp>0&&tmp<256){
      parsedline[j++]=tmp;
      i+=3;
     }
     E
      parsedline[j++]=line[i];
    }
    E
     parsedline[j++]=line[i];
   }
   /* handle hex truncation. */
   EI(line[i+2]&&isxdigit(line[i+1])&&isxdigit(line[i+2])){
    /* to not get confused with the length check wildcard. */
    IF(!(line[i+3]&&line[i+4]&&isdigit(line[i+3])&&isdigit(line[i+4]))){
     IF(isalpha(line[i+1]))
      tmp=(tolower(line[i+1])-0x56);
     E
      tmp=(line[i+1]-0x2F);
     tmp*=16;
     IF(isalpha(line[i+2]))
      tmp+=(tolower(line[i+2])-0x56);
     E
      tmp+=(line[i+2]-0x2F);
     tmp-=17;
     IF(tmp>0&&tmp<256){
      parsedline[j++]=tmp;
      i+=2;
     }
     E
      parsedline[j++]=line[i];
    }
    E
     parsedline[j++]=line[i];
   }
   /* handle random(alpha) truncation. */
   EI(line[i+1]==0x26){
    gettimeofday(&tv,(S timezone *)0);
    srand(tv.tv_usec);
    IF((2.0*rand()/(RAND_MAX+1.0))>1)
     parsedline[j++]=(0x41+(26.0*rand()/(RAND_MAX+1.0)));
    E
     parsedline[j++]=(0x61+(26.0*rand()/(RAND_MAX+1.0)));
    i++;
   }
   /* handle random(numeric) truncation. */
   EI(line[i+1]==0x23){
    gettimeofday(&tv,(S timezone *)0);
    srand(tv.tv_usec);
    parsedline[j++]=(0x30+(10.0*rand()/(RAND_MAX+1.0)));
    i++;
   }
   E
    parsedline[j++]=line[i];
  }
  E
   parsedline[j++]=line[i];
 }
 parsedline[j]=0x0;
 R;
}
/* this is suppost to get the input and output from the ruleset file. */
/* the file format is per line stacked, must have both.  like so: */
/* input given. (wildcards apply here) */
/* output to responed. */
/* input given. (wildcards apply here) */
/* output to responed. (and so on) */
/* also partly handles the "one time" variable. */
V makelists(C *path){
 /* yes, i could have reused some, but i like to be able to keep track. */
 /* loop ints. */
 U I i=0;
 U I j=0;
 U I k=0;
 U I l=0;
 /* rather large size for possible long requests, lower if desired. */
 U C rread[(BASE_BUFFER*16+1)];
 FL *fd;
 /* this switches to standard input if the argument is provided. */
 IF(inputrules){
  IF(!(fd=FRO(INPUTPATH,"r",stdin)))
   pe("standard input does not appear to exist or is not readable",1);
  pe("awaiting ruleset from standard input, followed by ^D",0);
 }
 E
  IF(!(fd=FO(path,"r")))
   pe("ruleset file does not appear to exist or is not readable",1);
 B(rread,SO(rread));
 W(FS(rread,SO(rread),fd)){
  /* maximum input/output storage met.  exits when it occurs. */
  IF(j>=MAX_ARGS)
   pe("too many slots to store in memory",1);
  IF(rread[0]!=0x23){
   F(i=0;i<SL(rread);i++)
    IF(rread[i]==0x0A)
     rread[i]=0x0;
   IF(rread[0]!=0x0A&&rread[0]!=0x0){
    /* simple on/off int to write line of input then line of output, then */
    /* checking at the end to make sure the input and output lines are */
    /* even. (at the end of the function) */
    IF(k)
     k=0;
    E
     k=1;
    IF(k){
     /* replaces the hex, dec and random variables. */
     parsecharvars(rread);
     IF(!(input[tot_i++]=(C *)SDU(parsedline)))
      pe("makelists(): duplication of memory error",1);
    }
    E{
     /* decided not to allow an only function sleep, so data is required. */
     IF(SL(rread)>6&&isdigit(rread[(SL(rread)-1)])&&isdigit(rread[(SL(rread)-
     2)])&&isdigit(rread[(SL(rread)-3)])&&isdigit(rread[(SL(rread)-4)])&&
     isdigit(rread[(SL(rread)-5)])&&rread[(SL(rread)-6)]==PARAMETER_VAR_CHAR){
      rs_delay[tot_o]=(((rread[(SL(rread)-5)]-0x30)*10000)+((rread[(SL(rread)-
      4)]-0x30)*1000)+((rread[(SL(rread)-3)]-0x30)*100)+((rread[(SL(rread)-2)]-
      0x30)*10)+(rread[(SL(rread)-1)]-0x30));
      rread[(SL(rread)-6)]=0x0;
     }
     /* this checks for the "one time" variable.  then, adds it to the int */
     /* buffer, if no data is provided after the variable it will be treated */
     /* as normal text. */
     IF(SL(rread)>SL(NS_ONCE)&&!SNCC(rread,NS_ONCE,SL(NS_ONCE))){
      /* kill the two variable bytes. (output) */
      F(l=SL(NS_ONCE);l<SL(rread);l++)
       rread[(l-SL(NS_ONCE))]=rread[l];
      rread[(l-SL(NS_ONCE))]=0x0;
      rs_static[tot_o]=1;
     }
     E
      rs_static[tot_o]=0;
     /* replaces the hex, dec and random variables. */
     parsecharvars(rread);
     IF(!(output[tot_o++]=(C *)SDU(parsedline)))
      pe("makelists(): duplication of memory error",1);
    }
   }
   B(rread,SO(rread));
  }
  j++;
 }
 FC(fd);
 /* check for eual amount of input to outputs, so the ruleset corresponds. */
 IF(tot_i!=tot_o)
  pe("input/output data from the ruleset does not correspond",1);
 /* no rules, no go. (might as well use telnet then) */
 IF(!tot_i||!tot_o)
  pe("no input/output responses supplied from the ruleset file",1);
 /* status showing for stdin ruleset input. */
 IF(inputrules)
  pe("valid ruleset supplied via input, continuing",0);
 R; 
}
/* displays argv[0] and the supplied text and exits if non-zero. */
V pe(C *err,U SH quit){
 /* simple enough handling on this option. */
 IF(!noshowa)
  P("%s: %s.\n",progname,err);
 IF(quit)
  nsexit(quit);
 R;
}
/* displays hex, dec or regular with the from/to prompt option. */
/* 0 = output, 1 = input, 2 = other. */
V pd(U C *in,U SH io,U I size){
 U I i=0;
 U I j=0;
 U I k=0;
 /* will be used for hex or dec (or duplication of the original). */
 C *chrdump;
 C *line;
 C prompt[4];
 /* columns gets set before pd() is ever ran, in main(). */
 IF(!(line=(C *)M(size+1)))
  pe("pd(): allocation of memory error",1);
 B(line,(size+1));
 IF(displaytype){
  /* size multiplied four times because every byte in the original */
  /* buffer is going to be equal to four in the new buffer. */
  IF(!(chrdump=(C *)M((SL(in)*4)+1)))
   pe("pd(): allocation of memory error",1);
  B(chrdump,SO(chrdump));
  F(i=0;i<SL(in);i++){
   /* 0 = character, 1 = hex, 2 = decimal. */
   IF(displaytype==1)
    SP(&chrdump[k],"\\x%.2x",in[i]);
   E
    SP(&chrdump[k],"%.3d%c",in[i],((i+1)!=SL(in))?0x20:0x0);
   k+=4;
  }
 }
 /* make the chrdump buffer a standard character buffer if no dispalytype to */
 /* wordwrap up later. */
 IF(!displaytype){
  IF(!(chrdump=(C *)SDU(in)))
   pe("pd(): duplication of memory error",1);
 }
 B(prompt,SO(prompt));
 IF(!io&&!nossend)
  SP(prompt,"-> ");
 EI(io&&!nosrecv)
  SP(prompt,"<- ");
 IF(!nowrap){
  B(line,(size+1));
  F(i=0;i<SL(chrdump);i++){
   IF(!SL(line))
    SP(line,"%s",!nofrom?prompt:"");
   /* next two checks are impossible to effect hex or dec dumps. */
   /* makes sure tabs do not screw up the wordwrap. */
   IF(chrdump[i]==0x09)
    /* 7 spaces = 1 tab(sort of, at a tab max), if it is beyond the buffer */
    /* size the tab will get cut off.  does not handle like a tab does. */
    F(j=0;(j<7&&!(SL(line)>=size));j++)
     line[SL(line)]=0x20;
   /* do not want a screwed up wordwrap, just slap a linefeed there instead. */
   EI(chrdump[i]==0x0A)
    line[SL(line)]=0x0D;
   E
    line[SL(line)]=chrdump[i];
   IF(SL(line)>=size){
    P("%s\n",line);
    B(line,(size+1));
   }
  }
  IF(SL(line))
   P("%s\n",line);
 }
 /* for disabled wordwrap. (-w option) */
 EI(SL(chrdump))
  P("%s%s\n",!nofrom?prompt:"",chrdump);
 FR(chrdump);
 FR(line);
 R;
}
/* this function does the logging. (-l option) */
/* 0 = output, 1 = input, 2 = other. */
V logdata(C *data,U SH io){
 /* will be used to hold the ctime data. */
 C *timebuf;
 C prompt[4];
 T tm;
 FL *fd;
 /* opens and allows appending to the log file. */
 IF(!(fd=FO(logfile,"a")))
  pe("log file can not be created or is not writable",1);
 E{
  time(&tm);
  IF(!(timebuf=(C *)SDU(ctime(&tm))))
   pe("logdata(): duplication of memory error",1);
  parameter(timebuf,3,0x20);
  FR(timebuf);
  B(prompt,SO(prompt));
  IF(!io)
   SP(prompt,"->");
  EI(io==1)
   SP(prompt,"<-");
  E
   SP(prompt,"!!");
  IF(SL(data))
   FP(fd,"%s %s %s\n",parm,prompt,data);
  FC(fd);
 }
 R;
}  
/* this is for use with dumping a programs output to the socket.  will not */
/* take direct to console data. */
V dumpexec(C *path,I s){
 U I i=0;
 U C eread[(BASE_BUFFER*4+1)];
 FL *efd;
 /* this is where netscript dumps non-stderr data to the socket from a */
 /* specified program. */ 
 IF(!noshowp)
  pe("entering execution dump, incoming data suspended",0);
 /* if the program isnt done after ?? seconds it is probably hung, kill it. */
 IF(xalrm)
  A(xalrm);
 IF(!(efd=popen(path,"r")))
  pe("execution dump failed, incoming data resumed",0);
 E{
  W(FS(eread,SO(eread),efd)){
   F(i=0;i<SL(eread);i++)
    IF(eread[i]==0x0A)
     eread[i]=0x0;
   IF(!nossend)
    pd(eread,0,columns);
   WR(sock,eread,SL(eread));
   WR(sock,"\n",1);
   /* add the output to the log, if enabled. (-l option) */
   IF(log)
    logdata(eread,0);
   B(eread,SO(eread));
  }
 }
 PC(efd);
 IF(xalrm)
  A(0);
 /* it would never make it to this point two times with variable usage. */
 /* so, no need to support such an event.  required information. */
  IF(runexit)
   pe("completed execution dump, exiting netscript",1);
  EI(!noshowp)
   pe("completed execution dump, incoming data resumed",0);
 R;
}
/* the socket connection and line parser.  takes the data from the remote */
/* host and puts it in a per line format, then passes it along to the match */
/* function(s).  also includes some specific variable handling. */
V sockconnect(C *host,U I port){
 /* many loop ints used to help keep track of what is what.  still re-used */
 /* some i knew would never cross paths. */
 U I i=0;
 U I j=0;
 U I k=0;
 U I l=0;
 U I m=0;
 U I n=0;
 U I o=0;
 U I p=0;
 SI I salen=0;
 /* file dump read buffer. (variable $<) */
 U C dread[(BASE_BUFFER*4+1)];
 /* special read buffer for initial data. */
 U C iread[(BASE_BUFFER*4+1)];
 /* initial write buffer. */
 U C iwrite[(BASE_BUFFER*4+1)];
 /* read buffer for socket data.  needs comparisons. */
 U C sread[(BASE_BUFFER*4+1)];
 /* buffer for data built up till a completed line, otherwise gets cleared. */
 /* allowance of large lines, lower if desired. (not lower than sread[]) */
 U C sreadl[(BASE_BUFFER*16+1)];
 /* three byte response buffer, for telnet protocol responses. */
 U C telnet[3];
 /* single byte, multiple uses.  no comparisons. */
 C sb[1];
 S hostent *he;
 /* connect to socket / bind client socket. */
 S SAN ns;
 /* bind server socket. */
 S SAN sns;
 FL *dfd;
 FL *ifd;
 IF(!noshowc)
  /* could have used tot_i or tot_o for the second argument, i used tot_i. */
  P("[%u] * Session started with %d rule(s) defined.\n",getpid(),tot_i);
 /* handling for socket binding. */
 IF(bindmode){
  /* main server sockaddr structure. */
  ssock=SK(AF_INET,SOCK_STREAM,IPPROTO_TCP);
  B((V *)&sns,SO(sns));
  sns.sin_family=AF_INET;
  sns.sin_port=htons(port);
  sns.sin_addr.s_addr=INADDR_ANY;
  /* bind to the port. (-b option) */
  IF(BI(ssock,(S SA *)&sns,SO(sns))<0){
   pe("could not bind properly",0);
   /* yet again.  better safe than sorry, do not want fds to fill up. */
   IF(sock){
    SD(sock,2);
    sock=CL(sock);
   }
   SD(ssock,2);
   ssock=CL(ssock);
   R;
  }
  IF(!noshowc)
   P("[%u] * Awaiting connection...\n",getpid());
  /* one time connection, so no need to backlog.  using zero. */
  IF(listen(ssock,0)<0){
   pe("could not listen on supplied port",0);
   R;
  }
  /* int p is only used here, once. */
  salen=SO(ns);
  IF((sock=accept(ssock,(S SA *)&ns,&salen))<0){
   pe("could not accept remote connection",0);
   R;
  }
  /* since bind is a one time use, close it up. */
  SD(ssock,2);
  ssock=CL(ssock);
  IF(!noshowc)
   P("[%u] * Accepted remote connection...\n",getpid());
  /* states log start time, if enabled. (-l option) */
  IF(log)
   logdata("NEW NETSCRIPT SESSION STARTED.",2);
 }
 /* handling for socket connection. */
 E{
  sock=SK(AF_INET,SOCK_STREAM,IPPROTO_TCP);
  ns.sin_family=AF_INET;
  ns.sin_port=htons(port);
  IF(!noshowc)
   P("[%u] * Checking the provided host...\n",getpid());
  IF(ns.sin_addr.s_addr=inet_addr(host)){ 
   /* if not valid, attempts to resolve as a host. */
   IF(!(he=gethostbyname(host))){
    /* nothing left to try. */
    pe("could not resolve the provided host",0);
    R;
   }
   E
    /* successfully resolved. */
    memcpy((C *)&ns.sin_addr,(C *)he->h_addr,SO(ns.sin_addr));
  }
  IF(!noshowc)
   P("[%u] * Attempting to connect to the provided host...\n",getpid());
  /* if you can not connect to the server the first time might as well kill */
  /* the program, same problem would continue infinitely.  so, if it times */
  /* out after ??(25 by default) seconds it will just exit. */
  IF(getenv(ENV_TIMEOUT)&&AT((C *)getenv(ENV_TIMEOUT))>0)
   A(AT((C *)getenv(ENV_TIMEOUT)));
  E
   A(25);
  /* like it looks, attempts to connect to the remote server. */
  IF(CO(sock,(S SA *)&ns,SO(ns))){
   pe("could not connect to the provided host",0);
   R;
  }
  E{
   A(0);
   IF(!noshowc)
    P("[%u] * Connected successfully...\n",getpid());
   /* states log start time, if enabled. (-l option) */
   IF(log)
    logdata("NEW NETSCRIPT SESSION STARTED.",2);
  }
 }
 /* cleans out the dynamic (int)buffer for the "one time" variable. */
 B((V *)rs_dynamic,SO(rs_dynamic));
 IF(initsend){
  IF(!SC(iswrite,"?")){
   IF(!(ifd=FRO(INPUTPATH,"r",stdin)))
    pe("could not open standard input for read",0);
   E{
    IF(!noshowp)
     pe("initial send defined user input, incoming data suspended",0);
    /* show default or reason prompt. */
    FP(stderr,"enter initial data to send: ");
    B(iread,SO(iread));
    /* truncated for pd(). */
    IF(FS(iread,SO(iread),ifd))
     F(o=0;o<SL(iread);o++)
      IF(iread[o]==0x0A)
       iread[o]=0x0;
    IF(!noshowp)
     pe("input received, incoming data resumed",0);
    IF(!nossend)
     pd(iread,0,columns);
    WR(sock,iread,SL(iread));
    WR(sock,"\n",1);
    IF(log)
     logdata(iread,0);
   }
   FC(ifd);
  }
  E{
   /* sends the initial data supplied by the argument if supplied. */
   IF(!nossend)
    pd(iswrite,0,columns);
   WR(sock,iswrite,SL(iswrite));
   WR(sock,"\n",1);
   /* add the initial data to the log, if enabled. (-i and -l option) */
   IF(log)
    logdata(iswrite,0);
  }
 }
 IF(runcmd)
  dumpexec(execfile,sock);
 IF(alrm)
  A(alrm);
 W(1){
  B(sread,SO(sread));
  IF(!RD(sock,sread,SO(sread))){
   pe("socket reading halted/finished",0);
   SD(sock,2);
   sock=CL(sock);
   R;
  }
  F(i=0;i<SL(sread);i++){
   /* telnet protocol handling, bytes set before handled. */
   /* using this method will have socket reading not wait for a \r or a \n. */
   /* (the -T option) */
   IF(tnet&&sread[i]==0xFF&&sread[i+1]&&sread[i+2]){
    B(telnet,SO(telnet));
    telnet[0]=0xFF;
    IF(sread[i+1]==0xFC||sread[i+1]==0xFB)
     telnet[1]=0xFE;
    EI(sread[i+1]==0xFE||sread[i+1]==0xFD)
     telnet[1]=0xFC;
    IF(telnet[1]==0xFE||telnet[1]==0xFC){
     /* checks to change socket handling for later. */
     IF(!truetnet){
      pe("received a telnet protocol option, entering telnet reading mode",0);
      truetnet=1;
     }
     /* the option. */
     telnet[2]=sread[i+2];
     WR(sock,telnet,SO(telnet));
     IF(!nossend&&!notnetopt)
      P("%s(TELNETD-OPT REPLIED: 0x%.2x 0x%.2x 0x%.2x -> 0x%.2x 0x%.2x 0x%.2x)"
      "\n",!nofrom?"-> ":"",sread[i],sread[i+1],sread[i+2],telnet[0],telnet[1],
      telnet[2]);
    }
   }
   /* telnet mode will not wait for a \r or \n, must get one or more telnet */
   /* option before automatically allowed. so, not to create problems with */
   /* bad syntax. */
   IF(sread[i]==0x0A||sread[i]==0x0D||(tnet&&truetnet&&!sread[i+1])){
    /* check where to palce the nullbyte and make sure byte exists for the */
    /* extended null byte.  it will not add the last byte if its the last */
    /* byte. */
    IF(tnet&&truetnet&&!sread[i+1]&&(BASE_BUFFER*16+1)>(j+1)){
     /* append the byte that is usually a \r or \n, but for telent support a */
     /* non-static character. */
     sreadl[j]=sread[i];
     sreadl[(j+1)]=0x0;
    }
    E
     sreadl[j]=0x0;
    IF(!nosrecv)
     pd(sreadl,1,columns);
    /* add the output to the log, if enabled. (-l option) */
    IF(log)
     logdata(sreadl,1);
    F(k=0;k<tot_i&&SL(sreadl);k++){
     IF(!stoprules&&!wordmatch(input[k],sreadl)&&!(rs_static[k]&&
     rs_dynamic[k])){
      /* cancel possibility of a repeated rule. */
      IF(rs_static[k]){
       IF(!noshowp)
        pe("ruleset defined this rule to only be used once",0);
       rs_dynamic[k]=1;
      }
      IF(rs_delay[k]){
       IF(!noshowp)
        pe("ruleset defined a delay before processing of this rule",0);
       SLP(rs_delay[k]);
      }
      /* parse the arguments input->output, $1 $2 $3, makes swrite buf. */
      parseoutput(sreadl,output[k]);
      /* check and handle the direct stop variable. */
      IF(!SNCC(swrite,NS_STOP,SL(NS_STOP))){
       /* changes buffer to dump the variable.  for later display. */
       F(p=SL(NS_STOP);p<SL(swrite);p++)
        swrite[(p-SL(NS_STOP))]=swrite[p];
       swrite[(p-SL(NS_STOP))]=0x0;
       /* no possibility of format bug, pe() handles like text. */
       IF(!noshowp)
        pe(SL(swrite)?swrite:"ruleset defined ruleset stop",0);
       stoprules=1;
      }
      /* check and handle for the direct quit variable. */
      EI(!SNCC(swrite,NS_QUIT,SL(NS_QUIT))){
       /* changes buffer to dump the variable.  for later display. */
       F(p=SL(NS_QUIT);p<SL(swrite);p++)
        swrite[(p-SL(NS_QUIT))]=swrite[p];
       swrite[(p-SL(NS_QUIT))]=0x0;
       /* no possibility of format bug, pe() handles like text. */
       IF(!noshowp)
        pe(SL(swrite)?swrite:"ruleset defined socket kill",0);
       SD(sock,2);
       sock=CL(sock);
       R;
      }
      /* check and handle the direct ask variable. */
      EI(!SNCC(swrite,NS_ASK,SL(NS_ASK))){
       /* changes buffer to dump the variable.  for later display. */
       /* little bit of repeating(sort of) code for the variable dumping, no */
       /* way around it without a larger function to handle it. (3 lines) */
       F(p=SL(NS_ASK);p<SL(swrite);p++)
        swrite[(p-SL(NS_ASK))]=swrite[p];
       swrite[(p-SL(NS_ASK))]=0x0;
       IF(!(ifd=FRO(INPUTPATH,"r",stdin)))
        pe("could not open standard input for read",0);
       E{
        IF(!noshowp)
         pe("ruleset defined user input, incoming data suspended",0);
        /* show default or reason prompt. */
        FP(stderr,"%s: ",SL(swrite)?swrite:"enter response");
        B(iwrite,SO(iwrite));
        IF(FS(iwrite,SO(iwrite),ifd))
         F(o=0;o<SL(iwrite);o++)
          IF(iwrite[o]==0x0A)
           iwrite[o]=0x0;
        IF(!noshowp)
         pe("input received, incoming data resumed",0);
        IF(!nossend)
         pd(iwrite,0,columns);
        WR(sock,iwrite,SL(iwrite));
        WR(sock,"\n",1);
        IF(log)
         logdata(iwrite,0);
       }
       FC(ifd);
      }
      /* check and handle the display variable. */
      EI(!SNCC(swrite,NS_ECHO,SL(NS_ECHO))){
       /* changes buffer to dump the variable.  for later display. */
       F(p=SL(NS_ECHO);p<SL(swrite);p++)
        swrite[(p-SL(NS_ECHO))]=swrite[p];
       swrite[(p-SL(NS_ECHO))]=0x0;
       /* no possibility of format bug, pe() handles like text. */
       /* ruleset defined this, so -C won't cut this off. */
       pe(SL(swrite)?swrite:"ruleset defined an empty display rule",0);
      }
      /* check and handle the no return write (raw) variable. */
      /* shows up like a standard write out. */
      EI(!SNCC(swrite,NS_RAW,SL(NS_RAW))){
       /* changes buffer to dump the variable.  for later display. */
       F(p=SL(NS_RAW);p<SL(swrite);p++)
        swrite[(p-SL(NS_RAW))]=swrite[p];
       swrite[(p-SL(NS_RAW))]=0x0;
       /* no possibility of format bug, pe() handles like text. */
       IF(!nossend)
        pd(swrite,0,columns);
       WR(sock,swrite,SL(swrite));
      }
      /* check and handle the file dump variable. */
      EI(!SNCC(swrite,NS_DUMP,SL(NS_DUMP))){
       /* changes buffer to dump the variable.  for later display. */
       F(p=SL(NS_DUMP);p<SL(swrite);p++)
        swrite[(p-SL(NS_DUMP))]=swrite[p];
       swrite[(p-SL(NS_DUMP))]=0x0;
       IF(access(swrite,R_OK))
        pe("could not access the file requested to dump in the ruleset",0);
       E{
        IF(!(dfd=FO(swrite,"r")))
         pe("file requested by ruleset does not appear to exist or is not read"
         "able",0);
        E{
         B(dread,SO(dread));
         /* no need to parse the data that comes from this file, instead it */
         /* directly dumps the file data.  single byte writing for all */
         /* character types. */
         IF(!noshowp)
          pe("ruleset requested a file dump to the remote host",0);
         W((n=FG(dfd))!=EOF){
          sb[0]=n;
          write(sock,sb,1);
         }
         IF(!noshowp)
          pe("ruleset requested file dump was successful",0);
        }
        FC(dfd);
       }
      }
      /* check and handle the write/append variable.  checks for append */
      /* first, beacuse if one variable is going to be larger it will be */
      /* the apppend variable. */
      EI(!SNCC(swrite,NS_WRITE,SL(NS_WRITE))||!SNCC(swrite,NS_APPEND,
      SL(NS_APPEND))){
       /* find out which variable it is, two in one function here. */
       IF(!SNCC(swrite,NS_APPEND,SL(NS_APPEND))){
        n=SL(NS_APPEND);
        /* append type, fopen(). */
        sb[0]=0x61;
       }
       E{
        n=SL(NS_WRITE);
        /* append write, fopen(). */
        sb[0]=0x77;
       }
       /* changes buffer to dump the variable.  for later display. */
       F(p=n;p<SL(swrite);p++)
        swrite[(p-n)]=swrite[p];
       swrite[(p-n)]=0x0;
       /* check for existance with non-writability. */
       IF(!access(swrite,F_OK)&&access(swrite,W_OK))
        pe("could not access the file requested to write to in the ruleset",0);
       E{
        IF(!(dfd=FO(swrite,sb)))
         pe("file requested to write to could not be opened properly",0);
        E{
         FP(dfd,"%s\n",sreadl);
         IF(!noshowp)
          pe("ruleset requested write out was successful",0);
        }
        FC(dfd);
       }
      }
      EI(!SNCC(swrite,NS_EXEC,SL(NS_EXEC))){
       /* changes buffer to dump the variable.  for later display. */
       F(p=SL(NS_ECHO);p<SL(swrite);p++)
        swrite[(p-SL(NS_ECHO))]=swrite[p];
       swrite[(p-SL(NS_ECHO))]=0x0;
       IF(!noshowp)
        pe("ruleset requested a local program execution",0);
       dumpexec(swrite,sock);
      }
      E{
       IF(!nossend)
        pd(swrite,0,columns);
       WR(sock,swrite,SL(swrite));
       WR(sock,"\n",1);
       /* add the input to the log, if enabled. (-l option) */
       IF(log)
        logdata(swrite,0);
      }
     }
    }
    B(sreadl,SO(sreadl));
    j=0;
   }
   E
    /* do not print the telnet options, no one wants to see that. */
    IF(tnet&&truetnet&&sread[i]==0xFF&&(sread[i+1]==0xFE||sread[i+1]==0xFD||
    sread[i+1]==0xFC||sread[i+1]==0xFB))
     i+=2;
    /* for reading of printable characters only.  meaning non-printable */
    /* characters will not make it to ruleset processing, but the telnet */
    /* protocol handling is processed before this point.  so, it will not be */
    /* effected. (-P option) */
    EI(!printonly||isprint(sread[i])){
     /* check the buffer that is filling until a completed line, making sure */
     /* not to exceed bounds.  if it does the buffer is cleared. */
     IF(j>=SO(sreadl)){
      B(sreadl,SO(sreadl));
      j=0;
     }
     /* write the byte if nothing before this option wanted to stop it. */
     E
      sreadl[j++]=sread[i];
    }
  }
 }
 /* once more.  better safe than sorry, do not want fds to fill up. */
 IF(sock){
  SD(sock,2);
  sock=CL(sock);
 }
 R;
}
/* for display types of version information. */
V showinfo(U SH type){
 C *timebuf;
 T tm;
 IF(type==1)
  P("[%u] * Version data: netscript/v%s!%s.\n"
  "[%u] * Id data: %s\n",getpid(),VERSION,COMPTYPE,getpid(),id);
 EI(type==2){
  time(&tm);
  IF(!(timebuf=(C *)SDU(ctime(&tm))))
   pe("main(): duplication of memory error",1);
  IF(timebuf[(SL(timebuf)-1)]==0x0A)
  timebuf[(SL(timebuf)-1)]=0x0;
  P("[ netscript: (version: v%s!%s - launchtime: %s) ]\n",VERSION,
  COMPTYPE,timebuf);
  FR(timebuf);
 }
 R;
}
/* display of variable uses, the -d argument. i used 0x9(tab) for this */
/* and the usage function in format form because if this file was */
/* copy/pasted it would lose its value and change its filesize. */
/* this is rather brief, better to consult documentation if present. */
V displayv(V){
 P("Defined ruleset match and replacement variables:\n"
 "IN/OUT%cTYPE%c%cDOES/MATCH%c%c%c%cVARIABLE\n"
 "O%cpre-data%cstops use of the ruleset%c%c\"%s\"\n"
 "O%cpre-data%ccloses the socket%c%c%c\"%s\"\n"
 "O%cpre-data%casks for input response to send%c%c\"%s\"\n"
 "O%cpre-data%cdisplays the supplied data%c%c\"%s\"\n"
 "O%cpre-data%cwrites the supplied data without a CR%c\"%s\"\n"
 "O%cpre-data%cuses rule only one time (requires data)%c\"%s\"\n"
 "O%cpre-data%cdumps the file used after variable%c\"%s\"\n"
 "O%cpre-data%cwrites to the file used after variable%c\"%s\"\n"
 "O%cpre-data%cappends to the file used after variable%c\"%s\"\n"
 "I%cwildcard%canything after point%c%c%c\"%s\"\n"
 "I%cwildcard%canything that is a value%c%c\"%s\"\n"
 "I%cwildcard%canything that is alphabetical%c%c\"%s\"\n"
 "I%cwildcard%canything that is numerical%c%c\"%s\"\n"
 "I%cwildcard%canything that is alphabetical|numerical%c\"%s\"\n"
 "I%cwildcard%canything that is upper case%c%c\"%s\"\n"
 "I%cwildcard%canything that is lower case%c%c\"%s\"\n"
 "I%cwildcard%canything that is punctuated%c%c\"%s\"\n"
 "I%cwildcard%canything that is a control charceter%c\"%s\"\n"
 "I%cwildcard%canything that is a printable word%c\"%s\"\n"
 "I%cwildcard%c<\"%c####\" anything that is equal to length, # = num>\n"
 "O%cend-data%c<\"%c#####\" delay in secs before processing rule, # = num>\n"
 "O%creplace%c%c<\"%c0-9\" will be replaced with the argument from input>\n"
 "I/O%creplace%c%c<\"%c01-FF\" will be replaced with hex->chr value>\n"
 "I/O%creplace%c%c<\"%c001-255\" will be replaced with dec->chr value>\n"
 "I/O%creplace%c%c<\"%c&\" will be replaced with a random letter>\n"
 "I/O%creplace%c%c<\"%c#\" will be replaced with a random number>\n",
 0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,NS_STOP,0x9,0x9,0x9,0x9,0x9,
 NS_QUIT,0x9,0x9,0x9,0x9,NS_ASK,0x9,0x9,0x9,0x9,NS_ECHO,0x9,0x9,0x9,NS_RAW,
 0x9,0x9,0x9,NS_ONCE,0x9,0x9,0x9,NS_DUMP,0x9,0x9,0x9,NS_WRITE,0x9,0x9,0x9,
 NS_APPEND,0x9,0x9,0x9,0x9,0x9,NS_ALL,0x9,0x9,0x9,0x9,NS_ANY,0x9,0x9,0x9,0x9,
 NS_ALPHA,0x9,0x9,0x9,0x9,NS_DIGIT,0x9,0x9,0x9,NS_ALNUM,0x9,0x9,0x9,0x9,
 NS_UPPER,0x9,0x9,0x9,0x9,NS_LOWER,0x9,0x9,0x9,0x9,NS_PUNCT,0x9,0x9,0x9,
 NS_CNTRL,0x9,0x9,0x9,NS_PRINT,0x9,0x9,PARAMETER_VAR_CHAR,0x9,0x9,
 PARAMETER_VAR_CHAR,0x9,0x9,0x9,PARAMETER_VAR_CHAR,0x9,0x9,0x9,
 PARAMETER_VAR_CHAR,0x9,0x9,0x9,PARAMETER_VAR_CHAR,0x9,0x9,0x9,
 PARAMETER_VAR_CHAR,0x9,0x9,0x9,PARAMETER_VAR_CHAR);
 nsexit(0);
}
/* stated in displayv(). */
V usage(U SH extended){
 P("Usage: %s [-i?t?l?e?x?XnNFTzHDwPcCAdv] -s|-r ruleset -b|-h host -p port\n"
 "Usage: %s ruleset (this must be an internal argument supplied ruleset)\n",
 progname,progname);
 IF(extended)
  P("\n  -r <ruleset>%c%cfile with input/output response data. (required or -s"
  ")\n"
  "  -h <host>%c%cremote host (or ip) to connect to. (required or -b)\n"
  "  -p <port>%c%cremote port to connect to or bind as. (required)\n"
  "  -i <data>%c%csend initial data once connected. (\"?\" for dynamic)\n"
  "  -t <timeout>%c%cclose the connection after alloted time.\n"
  "  -l <logfile>%c%cturns input/output logging on to the file.\n"
  "  -e <program>%c%cdumps the (non-direct) program data to the socket.\n"
  "  -x <timeout>%c%ckills netscript if the program dump hangs. (-e)\n"
  "  -X%c%c%ckills netscript after the program has been dumped. (-e)\n"
  "  -n%c%c%cdisables display of outgoing data.\n"
  "  -N%c%c%cdisables display of incoming data.\n"
  "  -F%c%c%cdisables display of arrow precursors.\n"
  "  -I%c%c%cwhen disconnected, reconnect to the remote host.\n"
  "  -s%c%c%ctakes the ruleset from standard input instead of a file.\n"
  "  -b%c%c%cbinds to the supplied port instead of connecting out.\n"
  "  -T%c%c%cattempts to interpret the telnet protocol.\n"
  "  -z%c%c%cdisables display of verbose telnetd-opt input/output.\n"
  "  -H%c%c%cdisplays hex values instead of character values. (ff)\n"
  "  -D%c%c%cdisplays dec values instead of character values. (255)\n"
  "  -w%c%c%cdisables wordwrap for in/out displaying.\n"
  "  -P%c%c%cdisables reading of non-printable characters.\n"
  "  -c%c%c%cdisables display of connection notifications.\n"
  "  -C%c%c%cdisables display of ruleset info that is not required.\n"
  "  -A%c%c%cdisables display of all prompted netscript comments.\n"
  "  -d%c%c%cdisplays the variables for use with the ruleset.\n"
  "  -v%c%c%cdisplays the version info of netscript and exits.\n"
  "  -V%c%c%cdisplays the version info of netscript and continues.\n"
  "\nAuthor: v9/vade79@realhalo.org, realhalo.org. [ihadnihn]\n",0x9,0x9,0x9,
  0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,
  0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,
  0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,
  0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9,0x9);
 nsexit(0);
}
/* any global exit routines can be placed here, your choice. */
V nsexit(SH i){
 exit(i);
}
/* program initializes here.  gets things started. */
I main(I argc,C **argv){
 U SH forever=0;
 U SH setfile=0;
 U SH sethost=0;
 U SH showv=0;
 U I i=0;
 U I j=0;
 U I k=0;
 U I l=0;
 U I sport=0;
 U I tshs=0;
 /* using the full maximum limit on arguments on this. */
 C *shost[MAX_ARGS];
 C *envcmd;
 C *rulesfile;
 T tm;
 /* just in case anyone would decide to run privileged, un-does such. */
 /* use to be setre*id(), but some platforms disagreed with it, this is ok. */
 IF(getuid()!=geteuid())
  seteuid(getuid());
 IF(getgid()!=getegid())
  setegid(getgid());
 IF(!(progname=(C *)SDU(argv[0])))
  pe("main(): duplication of memory error",1);
 /* ctrl-c. */
 SG(SIGINT,sighandler);
 /* segmentation fault, hope not to see that. */
 SG(SIGSEGV,sighandler);
 /* used for timed killing of netscript. */
 SG(SIGALRM,sighandler);
 /* other signals to catch, for cleanup. */
 SG(SIGILL,sighandler);
 SG(SIGTRAP,sighandler);
 SG(SIGBUS,sighandler);
 SG(SIGPIPE,sighandler);
 SG(SIGTERM,sighandler);
 /* +x ruleset file support, generalized support. */
 IF(argv[1]&&!argv[2]&&!access(argv[1],R_OK)&&!access(argv[1],X_OK))
  k=getexecargs(argv[1]);
 /* check the environment for the command line override variable.  or take */
 /* +x file arguments as env controller, all in one deal. */
 IF(getenv(ENV_CMDLINE)||k){
  /* priority 1 is +x files, priority 2 is env var and priority 3 is the */
  /* normal command line. */
  IF(k){
   IF(!(envcmd=M(SL(toenv)+SL(argv[0])+2)))
    pe("main(): allocation of memory error",1);
   B(envcmd,(SL(toenv)+SL(argv[0])+2));
  }
  E{
   IF(!(envcmd=M(SL((C *)getenv(ENV_CMDLINE))+SL(argv[0])+2)))
    pe("main(): allocation of memory error",1);
   B(envcmd,(SL((C *)getenv(ENV_CMDLINE))+SL(argv[0])+2));
  }
  IF(k){
   SP(envcmd,"%s %s",argv[0],toenv);
   /* done with that buffer. */
   FR(toenv);
  }
  E
   SP(envcmd,"%s %s",argv[0],(C *)getenv(ENV_CMDLINE));
  /* used this method to make a fresh argument list. */
  /* finding out how many elements are needed.  then, nulling the loop */
  /* routine since it was all done in the process itself. */
  F(argc=0;!parameter(envcmd,argc,0x20);argc++)
  0;
  /* creates the argument array.  also, added an extra slot for argv[0], */
  /* which will remain unused, but getopt() expects it. */
  IF(!(argv=(C **)M((argc+1)*(SO(C *)))))
   pe("main(): allocation of memory error",1);
  /* fill the array with the data from the env var or +x file. */
  /* (ENV_CMDLINE|+x) */
  F(argc=0;!parameter(envcmd,argc,0x20);argc++){
   IF(!(argv[argc]=(C *)SDU(parm)))
    pe("main(): duplication of memory error",1);
  }
  FR(envcmd);
 }
 /* interpret out the arguments supplied. */
 /* i changed a method here, due to me using shorts for most of these */
 /* options.  i did not want a possible loop over bug to occur if someone */
 /* was trying to break netscript with repeated arguments. (65536) */
 /* so, i am just doing int++.  avoids using another set of brackets. */
 W((i=getopt(argc,argv,"r:h:p:i:t:l:e:x:XnNFIsbTzHDwPcCAdvV"))!=EOF){
  SW(i){
   CS 'r':
    IF(setfile==0&&!setfile++)
     IF(!(rulesfile=(C *)SDU(optarg)))
      pe("main(): duplication of memory error",1);
    BR;
   CS 'h':
    IF(!bindmode)
     IF(sethost==0&&!sethost++)
      W(!parameter(optarg,tshs,0x2C)){
       IF(!(shost[tshs++]=(C *)SDU(parm)))
        pe("main(): duplication of memory error",1);
       IF(tshs>=MAX_ARGS)
        pe("too many slots to store in memory",1);
      }
    BR;
   CS 'p':
    IF(!sport)
     sport=AT(optarg);
    BR;
   CS 'i':
    IF(initsend==0&&!initsend++){
     /* parse out hex, dec and randoms. */
     parsecharvars(optarg);
     IF(!(iswrite=(C *)SDU(parsedline)))
      pe("main(): duplication of memory error",1);
    }
    BR;
   CS 't':
    IF(AT(optarg)>0&&!alrm)
     alrm=AT(optarg);
    BR;
   CS 'l':
    IF(log==0&&!log++)
     IF(!(logfile=(C *)SDU(optarg)))
      pe("main(): duplication of memory error",1);
    BR;
   CS 'e':
    IF(runcmd==0&&!runcmd++)
     IF(!(execfile=(C *)SDU(optarg)))
      pe("main(): duplication of memory error",1);
    BR;
   CS 'x':
    IF(AT(optarg)>0&&!xalrm)
     xalrm=AT(optarg);
    BR;
   CS 'X':
    runexit=1;
    BR;
   CS 'n':
    nossend=1;
    BR;
   CS 'N':
    nosrecv=1;
    BR;
   CS 'F':
    nofrom=1;
    BR;
   CS 'I':
    forever=1;
    BR;
   CS 's':
    inputrules=1;
    BR;
   CS 'b':
    IF(!sethost)
     bindmode=1;
    BR;
   CS 'T':
    tnet=1;
    BR;
   CS 'z':
    notnetopt=1;
    BR;
   CS 'H':
    /* 0 = character, 1 = hex, 2 = decimal. */
    displaytype=1;
    BR;
   CS 'D':
    /* 0 = character, 1 = hex, 2 = decimal. */
    displaytype=2;
    BR;
   CS 'w':
    nowrap=1;
    BR;
   CS 'P':
    printonly=1;
    BR;
   CS 'c':
    noshowc=1;
    BR;
   CS 'C':
    noshowp=1;
    BR;
   CS 'A':
    noshowa=1;
    BR;
   CS 'd':
    displayv();
    BR;
   CS 'v':
    showinfo(1);
    exit(0);
    BR;  
   CS 'V':
    showv=2;
    BR;  
   default:
    usage(0);
    BR;
  }
 }
 /* check for -r|-s, -h|-b and -p options. */
 IF((!setfile&&!inputrules)||(!sethost&&!bindmode)||(1>sport||65535<sport)){
  pe("no ruleset option, host|bind option and/or port option defined",0);
  usage(1);
 }
 /* main argument checking done, continuing on. */
 IF(showv)
  showinfo(showv);
 /* infinite connection or regular connection. */
 makelists(rulesfile); 
 /* this will stay if no environmental variable. */
 columns=DFL_COLUMNS;
 /* check for non-dfl columns, for the pd() display function with  wordwrap. */
 /* do this before using the pd() function at all. */
 IF(getenv(ENV_COLUMNS)){
  columns=AT((C *)getenv(ENV_COLUMNS));
  /* no need for anything larger unless youre tryingt to bother memory. */
  IF(1024<columns){
   pe("ENV_COLUMNS environmental variable is too large",0);
   columns=DFL_COLUMNS;
  }
  /* the prompt itself must be counted in the most minimal of cases. */
  IF((1>columns&&nofrom)||(4>columns&&!nofrom)){
   pe("ENV_COLUMNS environmental variable is too small",0);
   columns=DFL_COLUMNS;
  }
 }
 /* bind mode is for single usage, infinite connection attempts are for */
 /* automation of events. */
 IF(forever&&!bindmode)
  W(1){
   W(l<tshs){
    sockconnect(shost[l++],sport);
    /* better safe than sorry, do not want fds to fill up. */
    IF(sock){
     SD(sock,2);
     sock=CL(sock);
    }
    SLP(1);
   }
   l=0;
  }
 E{
  W(l<tshs||bindmode){
   sockconnect(bindmode?0:shost[l++],sport);
   /* turn bindmode off to leave this loop and exit. */
   bindmode=0;
   /* better safe than sorry, do not want fds to fill up. */
   IF(sock){
    SD(sock,2);
    sock=CL(sock);
   }
   SLP(1);
  }
 }
 nsexit(0);
}
/* for the real wildcard match checking. */
U SH wordmatch(C *wildcard,C *in){
 U I i=0;
 U I j=0;
 U I k=0;
 U I r=0;
 U I tmp=0;
 C *tmpbuf;
 /* remove back-to-back spaces for parsing with the remote data. */
 F(j=0;j<SL(in);j++)
  IF(!(in[j+1]&&in[j]==0x20&&in[j+1]==0x20))
   in[k++]=in[j];
 in[k]=0x0;
 /* remove back-to-back spaces for parsing with the ruleset. */
 F(k=j=0;j<SL(wildcard);j++)
  IF(!(wildcard[j+1]&&wildcard[j]==0x20&&wildcard[j+1]==0x20))
   wildcard[k++]=wildcard[j];
 wildcard[k]=0x0;
 /* quick exact match check, time saver. */
 IF(!SC(wildcard,in))
  R(0);
 W(!parameter(wildcard,i,0x20)){
  IF(!(tmpbuf=(C *)SDU(parm)))
   pe("wordmatch(): duplication of memory error",1);
  /* checks for a wildcard match for anything after that point.  before */
  /* checking the next parameter. */
  IF(SC(tmpbuf,NS_ALL)){
   /* gets the input parameter, to check for the rest of the wildcards or */
   /* possible perfect matches.  if no more parameters, returns failure. */
   /* also increases the word for the next loop run. */
   IF(parameter(in,i++,0x20)){
    FR(tmpbuf);
    R(1);
   }
   /* checking for all wildcard word matches. (length wildcard is next) */
   IF(!SC(tmpbuf,parm)||!SCC(tmpbuf,NS_ANY)||
   (!SCC(tmpbuf,NS_ALPHA)&&wischeck(parm,0))||
   (!SCC(tmpbuf,NS_DIGIT)&&wischeck(parm,1))||
   (!SCC(tmpbuf,NS_ALNUM)&&wischeck(parm,2))||
   (!SCC(tmpbuf,NS_UPPER)&&wischeck(parm,3))||
   (!SCC(tmpbuf,NS_LOWER)&&wischeck(parm,4))||
   (!SCC(tmpbuf,NS_PUNCT)&&wischeck(parm,5))||
   (!SCC(tmpbuf,NS_CNTRL)&&wischeck(parm,6))||
   (!SCC(tmpbuf,NS_PRINT)&&wischeck(parm,7)))
    r=0;
   /* the length wildcard, must be four digits long and then will be */
   /* processed accordingly. */
   EI(SL(tmpbuf)==5&&tmpbuf[0]==PARAMETER_VAR_CHAR&&isdigit(tmpbuf[1])&&
   isdigit(tmpbuf[2])&&isdigit(tmpbuf[3])&&isdigit(tmpbuf[4])){
    tmp=(((tmpbuf[1]-0x30)*1000)+((tmpbuf[2]-0x30)*100)+((tmpbuf[3]-0x30)*10)+
    (tmpbuf[4]-0x30));
    /* match. */
    IF(SL(parm)==tmp)
     r=0;
    /* did not match. */
    E
     r=1;
   }
   E
    r=1;
   IF(r){
    FR(tmpbuf);

    R(r);
   }
  }
  E{
   FR(tmpbuf);
   R(r);
  }
  FR(tmpbuf);
 }
 /* leftover data does not count as a match.  just using int i again for the
 /* loop.  checks to see if uneven arguments.  1 = failed, 0 = success. */
 i=0;
 W(!parameter(wildcard,i,0x20))
  i++;
 IF(!parameter(in,i,0x20))
  /* un-even. */
  r=1;
 E
  /* even. */
  r=0;
 R(r);
}
/* this is meant to pluck a word out of a sentence, then duplicated it to */
/* the character array "parm". */
U SH parameter(C *array,I i,U I sep){
 U I j=0;
 U I k=0;
 U I r=0;
 C *buf;
 IF(!(buf=(C *)M(SL(array)+1)))
  pe("parameter(): allocation of memory error",1);
 B(buf,(SL(array)+1));
 IF(i<0)
  r=1;
 E
  F(j=0;j<SL(array);j++){
   IF(array[j]==sep)
    i--;
   EI(!i)
    buf[k++]=array[j];
   IF(array[j]==0x0A||array[j]==0x0)
    j=(SL(array)+1);
  }
 IF(i>0)
  r=1;
 FR(parm);
 IF(!(parm=(C *)SDU(buf)))
  pe("parameter(): duplication of memory error",1);
 FR(buf);
 R(r);
}
/* the next functions checks word format of their counterpart, instead of a */
/* single character.  for some reason i like using strlen() instead of *buf */
/* to check for use, old habit i suppose. */
U SH wischeck(C *word,U I type){
 U I i=0;
 IF(!SL(word))
  R(0);
 F(i=0;i<SL(word);i++)
  IF((!isalpha(word[i])&&type==0)||(!isdigit(word[i])&&type==1)||
  (!isalnum(word[i])&&type==2)||(!isupper(word[i])&&type==3)||
  (!islower(word[i])&&type==4)||(!ispunct(word[i])&&type==5)||
  (!iscntrl(word[i])&&type==6)||(!isprint(word[i])&&type==7))
   R(0);
 R(1);
}
/* gets arguments out of +x files, if no arguments it will return false and */
/* not continue. */
U SH getexecargs(C *ruleset){
 /* not quite as big as the storage buffer, due the the fact this is just */
 /* for argument storage.  toenv will get freed when done with in main(). */
 U I i=0;
 U I j=0;
 /* if it is bigger than that someone is just trying to be evil. */
 U C rread[(BASE_BUFFER*4+1)];
 FL *fd;
 IF(!(fd=FO(ruleset,"r")))
  pe("ruleset file does not appear to exist or is not readable",1);
 B(rread,SO(rread));
 /* only need to read the first two lines for the args. */
 W(FS(rread,SO(rread),fd)&&j<2){
  IF(SL(rread)>2&&!SNC(rread,"#?",2)){
   /* ?(args) + -2(#?) + ?(ruleset) + 4(-r,space,space) + 1(extra). */
   IF(!(toenv=M(SL(rread)-2+SL(ruleset)+4+1)))
    pe("getexecargs(): allocation of memory error",1);
   B(toenv,(SL(rread)-2+SL(ruleset)+4+1));
   SP(toenv,"-r %s ",ruleset);
   F(i=2;i<SL(rread);i++){
    /* finish true, and append 0x0. */
    IF(rread[i]==0x0A){
     toenv[SL(toenv)]=0x0;
     FC(fd);
     R(1);
    }
    toenv[SL(toenv)]=rread[i];
   }
  }
  j++;
 }
 FC(fd);
 R(0);
}
/* EOF */
